

################################################################################


class _MagnetostaticField:
  

  _bemStatic = None
  _maxSteps_gmres = 4000
  
  
#-----------------------------------------------------------------------------#


  def __init__(self, mesh, other_cf, solvetol):
    from ngsolve import GridFunction, VectorH1, LinearForm, SymbolicLFI, BVP
    from . import _detailsMagnetostaticField
    
    self._NeumannSpace = type(self)._discreteSpace(mesh, order=1)
    self._DirichletSpace = type(self)._discreteSpace(mesh \
      , order=1, dirichlet="bc_dirichlet")

    self.comp1 = GridFunction(self._NeumannSpace)
    self.comp2 = GridFunction(self._DirichletSpace)

    self._gf = GridFunction(VectorH1(mesh, order=1))
    self.cf = type(self)._decomposition(self.comp1, self.comp2)

    phi1 = self._NeumannSpace.TestFunction()
    self._f1 = LinearForm(self._NeumannSpace)
    self._f1 += SymbolicLFI((type(self)._lfOperator(phi1)*other_cf \
      ).Compile(realcompile=True, wait=True))

    if (type(self)._bemStatic == None):
      type(self).__base__._bemStatic = \
        _detailsMagnetostaticField.DetailsMagnetostaticField(mesh \
        , self._NeumannSpace, self._DirichletSpace, self.comp2, solvetol)

    if (type(self)._femStatic == None):
      type(self)._femStatic = type(self)._femInitializer(mesh \
        , self._NeumannSpace, self._DirichletSpace, self.comp2, solvetol)

    self.bvp2 = BVP(bf=type(self)._femStatic.a2, lf=type(self)._femStatic.f2 \
      , gf=self.comp2, pre=type(self)._femStatic.c2 \
      , maxsteps=type(self)._maxSteps_gmres, prec=solvetol)

    self.Compute(solvetol)
    
  
#-----------------------------------------------------------------------------#


  def SetFromOther(self, other):
    self.comp1.vec.FV().NumPy()[:] = other.comp1.vec.FV().NumPy()
    self.comp2.vec.FV().NumPy()[:] = other.comp2.vec.FV().NumPy()
  
  
#-----------------------------------------------------------------------------#


  def SetZero(self):
    self.comp1.vec.FV().NumPy()[:] = 0.0
    self.comp2.vec.FV().NumPy()[:] = 0.0
  
  
#-----------------------------------------------------------------------------#


  def Compute(self, solvetol):
    from ngsolve.ngstd import Timer
    from ngsolve import BND

    magnetostaticTimer_comp1 = Timer("magnetostaticField::comp1")
    magnetostaticTimer_g = Timer("magnetostaticField::g")
    magnetostaticTimer_comp2 = Timer("magnetostaticField::comp2")
    
    magnetostaticTimer_comp1.Start()
    self._f1.Assemble()
    self.comp1.vec.data = self._femStatic.solverComp1 * self._f1.vec
    magnetostaticTimer_comp1.Stop()
    
    magnetostaticTimer_g.Start()
    g = self._bemStatic.bemMagnetostaticComputeBCs(self.comp1 \
      , self._NeumannSpace, solvetol)
    magnetostaticTimer_g.Stop()
    
    magnetostaticTimer_comp2.Start()
    self.comp2.Set(g, BND)
    self.bvp2.Do()
    magnetostaticTimer_comp2.Stop()


#-----------------------------------------------------------------------------#


  def _SetGF(self):
    for d in range(3):
      self._gf.components[d].Set(self.cf[d])

    
#-----------------------------------------------------------------------------#


  def Save(self, filepath):
    from commix._tools import SaveMagnetization

    self._SetGF()
    SaveMagnetization(filepath, self._NeumannSpace.mesh, self._gf \
      , label=type(self)._label)


#-----------------------------------------------------------------------------#


################################################################################
