

################################################################################


class Derivative:
  # represents the derivative of a magnetization
  
  
#-----------------------------------------------------------------------------#


  def __init__(self, mesh, V, X, maxwellCoupling, solvetol, j_cf):
    from ngsolve import GridFunction
    from . import strayField, oerstedField
    # X is the compound FESpace \subset H1(\Omega)^3
    self._mesh = mesh
    
    self.maxwellCoupling = maxwellCoupling
    
    self.gf = GridFunction(X)
    self.Dgf = self.gf.Deriv()
    self.cf = self.gf
    
    if( self.maxwellCoupling == "magnetostaticField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    if( self.maxwellCoupling == "strayField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
    elif( self.maxwellCoupling == "oerstedField" ):
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
  

#-----------------------------------------------------------------------------#


  def UpdateMaxwellCoupling(self, newMaxwellCoupling, solvetol, j_cf):
    from . import strayField, oerstedField

    self.maxwellCoupling = newMaxwellCoupling
    self.hs = None
    self.hc = None

    if( self.maxwellCoupling == "magnetostaticField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    elif( self.maxwellCoupling == "strayField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
    elif( self.maxwellCoupling == "oerstedField" ):
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    elif( self.maxwellCoupling == "eddyCurrent" ):
      pass
    elif( self.maxwellCoupling != "none" ): 
      print("Error in ._Mag.__init__():")
      print("Maxwell options not >strayField<, >oerstedField<, >eddyCurrent< \
        or >none< are not implemented yet")
      sys.exit()
    print("After Maxwell")

    
#------------------------------------------------------------------------------#


  def ComputeMaxwell(self, solvetol):
    import sys

    if(self.maxwellCoupling == "magnetostaticField" ):
      self._ComputeStrayField(solvetol)
      self._ComputeOerstedField(solvetol)
    elif(self.maxwellCoupling == "strayField"):
      self._ComputeStrayField(solvetol)
    elif(self.maxwellCoupling == "oerstedField"):
      self._ComputeOerstedField(solvetol)
    elif(self.maxwellCoupling != "none"):
      print("Error in _Mag.v.ComputeMaxwell:")
      print("maxwellCoupling =", self.maxwellCoupling, "unknown.")
      sys.exit()
  

#------------------------------------------------------------------------------#

  
  def _ComputeStrayField(self, solvetol):
    import sys
    if ( self.maxwellCoupling != "strayField" \
      and self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag.v._ComputeStrayField():")
      print("maxwellCoupling != <strayField> !! --> Terminate.")
      sys.exit()
    
    self.hs.Compute(solvetol)
  
  
#-----------------------------------------------------------------------------#

  
  def _ComputeOerstedField(self, solvetol):
    import sys
    if ( self.maxwellCoupling != "oerstedField" \
      and self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag.v._ComputeOerstedField():")
      print("maxwellCoupling != <oerstedField> !! --> Terminate.")
      sys.exit()
    
    self.hc.Compute(solvetol)
  
  
#-----------------------------------------------------------------------------#


  def phiUniaxialAnisotropy(self, e):
    from commix._tools.algebra import InnerProd
    return -1.0 * InnerProd(e, self.gf) * e
  
  
#-----------------------------------------------------------------------------#


  def opZhangLi(self, vel):
    return (sum(vel[k]*self.Dgf[0,k] for k in range(3)), \
      sum(vel[k]*self.Dgf[1,k] for k in range(3)), \
      sum(vel[k]*self.Dgf[2,k] for k in range(3)) )


#-----------------------------------------------------------------------------#


################################################################################
