#from . import boundingBox # TOECDO


################################################################################


class Geometry:
  
  def __init__(self, geometry=None, meshSize=None, scaling=1.0):
    
    self._geometry = geometry
    self.meshSize = meshSize
    self.scaling = scaling
    
#TOECDO    self.boundingBox = None

    self._mesh = None
    

#------------------------------------------------------------------------------#

  
  def GetReady(self):
    import sys
    from netgen.csg import CSGeometry
    from ngsolve import Mesh as ngsolveMesh
    from netgen.meshing import Mesh as netgenMesh
    from commix._setup.geometries.standardGeometries import _standardGeometry

    if not(self._mesh == None):
      print("From Geometry.GetReady(): Nothing to do, mesh is ready.")
      return

    
    # TOECDO
#    if (self.boundingBox == None):
#      self.boundingBox = self
#    else:
#      self._mesh = self.boundingBox.GetReady(self._geometry, self.meshSize \
#        , self.scaling)
#      return

    if(isinstance(self._geometry, str)):
      if(self.meshSize != None):
        print("Error in Geometry.GetReady():")
        print("meshSize must not be set, if mesh is loaded from file!!!")
        print(" ---> Terminate.")
        sys.exit()

      ngmesh = netgenMesh(3)
      ngmesh.Load( self._geometry )
    else:
      if( isinstance(self._geometry, _standardGeometry._StandardGeometry)):
        ngGeo = self._geometry.GetNgSolveGeometry()
      else:
        ngGeo = self._geometry

      if(self.meshSize == None):
        print("Error in Geometry.GetReady():")
        print("Specify meshSize --> Terminate.")
        sys.exit()
    
      geo = CSGeometry()
      geo.Add(ngGeo)
      
      ngmesh = geo.GenerateMesh( maxh = self.meshSize )
      
    self._mesh = ngsolveMesh( ngmesh )
  
  
#------------------------------------------------------------------------------#
# TOECDO
#  
#  def IsBoundingBox(self):
#    return False
#  
#  
#------------------------------------------------------------------------------#
# TOECDO
#  
#  def Get_restrictionMap(self):
#    return range(0,self._mesh.nv)
#  
#  
#------------------------------------------------------------------------------#
# TOECDO
#  
#  def Get_inner_domain_nr(self):
#    import sys
#    print("Error in Geometry.Get_inner_domain_nr():")
#    print("self.IsBoundingBox() = ", self.IsBoundingBox())
#    print("Something went wrong with the meshes:")
#    print("This method should never be called!")
#    sys.exit()
#  
#  
#------------------------------------------------------------------------------#
# TOECDO
#  
#  def Get_outer_domain_nr(self):
#    import sys
#    print("Error in Geometry.Get_outer_domain_nr():")
#    print("self.IsBoundingBox() = ", self.IsBoundingBox())
#    print("Something went wrong with the meshes:")
#    print("This method should never be called!")
#    sys.exit()
#  
#  
#------------------------------------------------------------------------------#

  
  def __str__(self):
    string = "\n\n ----- " + self.__class__.__name__ + " ----- \n\n"
    if not(self._mesh == None):
      string += "Vertices: " + str(self._mesh.nv) + "\n"
      string += "Elements: " + str(self._mesh.ne) + "\n"
      string += "meshSize = " + str(self.meshSize) + "\n"
      string += "scaling = " + str(self.scaling) +"\n"

      if isinstance(self._geometry, str):
        string += "file = " + str(self._geometry) + "\n"
# TOECDO
#      if ( self.boundingBox.IsBoundingBox() ):
#        string += self.boundingBox.__str__()

      return string
    else:
      string = "From Geometry.__str__(): Mesh not ready: Call GetReady() first!"
  
    return string

  
#------------------------------------------------------------------------------#

  
  def GetMesh(self):
    if not(self._mesh == None):
      return self._mesh
    else:
      print("From Geometry.GetMesh() : \
        Mesh not ready: Call GetReady() first!")
  
  
#------------------------------------------------------------------------------#

  
  def Save(self, filename):
    if not(self._mesh == None):
      self._mesh.ngmesh.Save( filename )
# TOECDO
#      if ( self.boundingBox.IsBoundingBox() ):
#        tmp = filename.rfind("/")+1
#        self.boundingBox.Save( filename[:tmp]+"boundingBox_"+filename[tmp:] )
    else:
      print("From Geometry.Save(): \
        Mesh not ready: Call GetReady() first!")
  
  
#------------------------------------------------------------------------------#


################################################################################
