### class Parameters ###


################################################################################


class Parameters:
  '''
  Objects of this class are meant to specify
    - material parameters [in SI units]
    - physical constants [those are predefined]
    - which lower-order contributions to consider:
      .H_ext = [ (f,g,h) where f, g, h are either constants 
                 or functions in (t, x, y, z) ]
      .maxwellCoupling = ['none', 'magnetostaticField', 'strayField','oerstedField']
      (magnetostaticField = strayfield & oersted field)
      .dmCoupling = ['none', 'bulk', 'interfacial']
      .spintronicsCoupling = ['none', 'zhangLi', 'slonczewski']
  '''
  

  def __init__(self, A=None, Ms=None, K=0.0, D=0.0, dmi_axis='z', gamma0=2.212761569e+05, mu0=1.2566370614359173e-06, alpha=1.0, T_start=None, T_end=None, timeStepSize=None, H_ext=None, easy_axis=(1,0,0), g=2.0023193043617, muB=9.27400968e-24, e=1.6021766208e-19, P=None, xi=None, dmCoupling="none", maxwellCoupling="strayField", spintronicsCoupling="none", hbar=1.054571800e-34, Je=None, d=None, p=(1,0,0), TOL=1.0e-8, preconditionerType="jacobi", QmatType="householder", QmatAxis="e3", guessType="old", theta=0.5, m0=None):
    self.A = A
    self.Ms = Ms
    self.K = K
    self.D = D
    self.dmi_axis = dmi_axis

    self.gamma0 = gamma0
    self.mu0 = mu0
    
    self.alpha = alpha

    self.T_start = T_start
    self.T_end = T_end
    self.timeStepSize = timeStepSize
    
    self.H_ext = H_ext
    self.easy_axis = easy_axis
    
    self.dmCoupling = dmCoupling
    self.maxwellCoupling = maxwellCoupling
#    self.sigma_in = 1.0 #TOECDO
#    self.sigma_out = None #TOECDO
    
    # spintronics par
    self.spintronicsCoupling = spintronicsCoupling
    self.Je = Je
    self.e = e
    # zhangLi par
    self.P = P
    self.g = g
    self.muB = muB
    self.xi = xi
    # slonczewski par
    self.hbar = hbar
    self.d = d
    self.p = p

    self.TOL = TOL
    self.preconditionerType = preconditionerType
    self.guessType = guessType
    self.QmatType = QmatType
    self.QmatAxis = QmatAxis

    self.theta = theta

    self.m0 = m0


#------------------------------------------------------------------------------#


  def __str__(self):
    string = "\n\n ----- " + self.__class__.__name__ + " ----- \n\n"
    for key, value in self.__dict__.items():
      if not(key.startswith("_")):
        string += key + " = " + str(value) + "\n"
    
    return string


################################################################################
