### TPS tmeplate ###
from . import _integrator


################################################################################


class TPSX(_integrator._Integrator):
  

#------------------------------------------------------------------------------#

  
  def GetReady(self):

    _integrator._Integrator.GetReady(self)


#------------------------------------------------------------------------------#


  def _SetUpSchemeSpecificObjects(self):

    self._SetQmatTransformationType(self.parameters.QmatType)
    self._SetQmatAxis(self.parameters.QmatAxis)


#------------------------------------------------------------------------------#


  def _StepPrepareWithTaskManager(self):

    self._ComputeMaxwell()
    self._AssembleSystem()


#------------------------------------------------------------------------------#


  def _AssembleLHS(self):
    self._Assemble_A_instat()


#------------------------------------------------------------------------------#


  def _StepPrepareWithoutTaskManager(self):

    self._Compute_Q()
    self._ComputeLHS()
    self._ComputeRHS()
    self._UpdatePreconditioner()


#------------------------------------------------------------------------------#


  def _ComputeLHS(self):
    self._SetLHS_3D()
    self._LHS = self._Q.dot(self._LHS_3D.dot(self._Q.transpose()))


#------------------------------------------------------------------------------#


  def _SetLHS_3D(self):
    self._LHS_3D = self._A_stat + self._A_instat


#------------------------------------------------------------------------------#


  def _ComputeRHS(self):
    self._SetRHS_3D()
    self._RHS = self._Q.dot(self._RHS_3D)


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner(self):
    if(self.parameters.preconditionerType == "jacobi"):
      self._UpdatePreconditioner_Jacobi()
    elif(self.parameters.preconditionerType == "practical"):
      self._UpdatePreconditioner_Practical()
    elif(self.parameters.preconditionerType == "stationary"):
      self._UpdatePreconditioner_Stationary()
    elif(self.parameters.preconditionerType == None):
      self._UpdatePreconditioner_None()
    else:
      raise ValueError("preconditionerType = ", self.parameters.preconditionerType \
        , "not known")


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_Jacobi(self):
    qjacq = 1.0 / self._LHS_3D.diagonal()[:self._LHS_3D.shape[0]//3]
    self._preJacobi = qjacq.repeat(2)


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_Practical(self):
    pass


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_Stationary(self):
    pass


#------------------------------------------------------------------------------#


  def _UpdatePreconditioner_None(self):
    pass


#------------------------------------------------------------------------------#


  def _UpdateGuess(self):
    if(self.parameters.guessType == "old"):
      self._UpdateGuess_Old()
    elif(self.parameters.guessType == None):
      self._UpdateGuess_None()
    else:
      raise ValueError("guessType = ", self.parameters.guessType, "not known")


#------------------------------------------------------------------------------#


  def _UpdateGuess_Old(self):
    self._guess = self._solution


#------------------------------------------------------------------------------#


  def _UpdateGuess_None(self):
    self._guess = None


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner(self):
    if(self.parameters.preconditionerType == "jacobi"):
      self._SetUpPreconditioner_Jacobi()
    elif(self.parameters.preconditionerType == "practical"):
      self._SetUpPreconditioner_Practical()
    elif(self.parameters.preconditionerType == "stationary"):
      self._SetUpPreconditioner_Stationary()
    elif(self.parameters.preconditionerType == None):
      self._SetUpPreconditioner_None()
    else:
      raise ValueError("preconditionerType = ", self.parameters.preconditionerType \
        , "not known")


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_Jacobi(self):
    import scipy.sparse.linalg
    qjacq = 1.0 / self._A_stat.diagonal()[:self._A_stat.shape[0]//3]
    self._preJacobi = qjacq.repeat(2)
    
    jacobi2D = lambda x: self._preJacobi*x
    self._preconditioner = scipy.sparse.linalg.LinearOperator( \
      (self._preJacobi.size, self._preJacobi.size), matvec=jacobi2D)


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_Practical(self):
    import scipy.sparse.linalg
    import scipy.sparse as sp
    import numpy as np
    A = self._A_stat[:self._A_stat.shape[0]//3, :self._A_stat.shape[0]//3]

    Ainv = np.zeros(A.shape, dtype=float)

    e = np.zeros(A.shape[0], dtype=float)
    e[0] = 1.0

    print("Inverting matrix for practical preconditioner")
    sol, succ = scipy.sparse.linalg.cg(A, e \
      , tol=self._solvetol, maxiter=4000)
    idx = (abs(sol) > self._solvetol)
    Ainv[idx,0] = sol[idx]
    for j in range(1, A.shape[0]):
      e[j-1] = 0.0
      e[j] = 1.0
      sol, succ = scipy.sparse.linalg.cg(A, e, tol=self._solvetol, maxiter=4000)
      idx = (abs(sol) > self._solvetol)
      Ainv[idx, j] = sol[idx]
      if (j % 100 == 0):
        print("Finished ", j, "out of ", A.shape[0] \
          , "cols of practical preconditioner")

    N = A.shape[0]
    practical3D = np.zeros((N*3, N*3), dtype=float)
    practical3D[:N, :N] = Ainv
    practical3D[N:2*N, N:2*N] = Ainv
    practical3D[2*N:3*N, 2*N:3*N] = Ainv

    self._prePractical = sp.csr_matrix(practical3D)
    self._prePractical.eliminate_zeros()
    
    self._Compute_Q()
    practical2D = lambda x: \
      self._Q.dot(self._prePractical.dot(self._Q.transpose().dot(x)))

    self._preconditioner = scipy.sparse.linalg.LinearOperator( \
      (self._Q.shape[0], self._Q.shape[0]), matvec=practical2D)


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_Stationary(self):
    import scipy.sparse.linalg
    import scipy.sparse as sp
    import numpy as np
    A = self._A_stat[:self._A_stat.shape[0]//3, :self._A_stat.shape[0]//3]

    Ainv = np.zeros(A.shape, dtype=float)

    e = np.zeros(A.shape[0], dtype=float)
    e[0] = 1.0

    print("Inverting matrix for practical preconditioner")
    sol, succ = scipy.sparse.linalg.cg(A, e \
      , tol=self._solvetol, maxiter=4000)
    idx = (abs(sol) > self._solvetol)
    Ainv[idx,0] = sol[idx]
    for j in range(1, A.shape[0]):
      e[j-1] = 0.0
      e[j] = 1.0
      sol, succ = scipy.sparse.linalg.cg(A, e, tol=self._solvetol, maxiter=4000)
      idx = (abs(sol) > self._solvetol)
      Ainv[idx, j] = sol[idx]
      if (j % 100 == 0):
        print("Finished ", j, "out of ", A.shape[0] \
          , "cols of practical preconditioner")

    N = A.shape[0]

    cooAinv = sp.coo_matrix(Ainv)
    self._preStationary = sp.csr_matrix(( \
      np.append(cooAinv.data, cooAinv.data), ( \
      np.append(2*cooAinv.row, 2*cooAinv.row+1), \
      np.append(2*cooAinv.col, 2*cooAinv.col+1) )))
      
    self._preStationary.eliminate_zeros()
    
    stationary = lambda x: self._preStationary.dot(x)

    self._preconditioner = scipy.sparse.linalg.LinearOperator( \
      self._preStationary.shape, matvec=stationary)


#------------------------------------------------------------------------------#

  
  def _SetUpPreconditioner_None(self):
    self._preconditioner = None


#------------------------------------------------------------------------------#

  
  def _SetQmatTransformationType(self, trafoType):
    import sys

    if trafoType == "householder":
      self._Q_proj = self._H_proj
    elif trafoType == "rotation":
      self._Q_proj = self._R_proj
    else:
      print("From _SetQmatTransformationType: parameters.QmatType =", trafoType)
      print("unknown --> Terminate")
      sys.exit()


#------------------------------------------------------------------------------#

  
  def _SetQmatAxis(self, axis):
    import sys

    if axis == "e1":
      self._Q_axisT = self._Q_axisTe1
    elif axis == "e2":
      self._Q_axisT = self._Q_axisTe2
    elif axis == "e3":
      self._Q_axisT = self._Q_axisTe3
    else:
      print("From _SetQmatAxis: parameters.QmatAxis =", axis)
      print("unknown --> Terminate")
      sys.exit()


#------------------------------------------------------------------------------#


  @classmethod
  def _Q_axisTe1(cls, q):
    if len(q) == 3:
      return [q[2], q[1], q[0]]

    return [q[4], q[5], q[2], q[3], q[0], q[1]]


#------------------------------------------------------------------------------#


  @classmethod
  def _Q_axisTe2(cls, q):
    if len(q) == 3:
      return [q[0], q[2], q[1]]

    return [q[0], q[1], q[4], q[5], q[2], q[3]]


#------------------------------------------------------------------------------#


  @classmethod
  def _Q_axisTe3(cls, q):
    return q


#------------------------------------------------------------------------------#

  
  def _SetUpAuxiliaryFunctions(self):
    from commix._gridFunctions import Derivative
    if self.parameters.spintronicsCoupling == "slonczewski":
      jprime_cf = self._jprime_cf*self._p_cf if self._jprime_cf != None else \
        self._jprime_cf
      self._Mag.v = Derivative(self._GetMesh(), self._V, self._X \
        , self.parameters.maxwellCoupling, self._solvetol, jprime_cf)
    else:
      self._Mag.v = Derivative(self._GetMesh(), self._V, self._X \
        , self.parameters.maxwellCoupling, self._solvetol, self._jprime_cf)

    self._Q = None
    self._Q_proj = None
    self._Q_axisT = None
    self._preJacobi = None
    self._prePractical = None
    self._preStationary = None


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):

    _integrator._Integrator._UpdateMaxwellCoupling(self)

    if self.parameters.spintronicsCoupling == "slonczewski":
      self._Mag.v.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
        , self._solvetol, self._jprime_cf * self._p_cf)
    else:
      self._Mag.v.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
        , self._solvetol, self._jprime_cf)


#------------------------------------------------------------------------------#


  def _StepSolve(self):
    self._Compute_v()


#------------------------------------------------------------------------------#


  def _Assemble_A_instat(self):
    self._a_instat.Assemble()
    self._A_instat.data = self._a_instat.mat.COO()[2].NumPy()


#------------------------------------------------------------------------------#


  def _AssembleRHS(self):
    import sys
    print("From _tpsx.AssembleRHS():")
    print("I'm only the base class, I don't know what to do.")
    print("--> Terminate")
    sys.exit()


#------------------------------------------------------------------------------#


  def _SkewTerm(self):
    return self._Skew()


#------------------------------------------------------------------------------#


  def _ExchangeTerm(self):
    return -self._C_ex * self._Exchange()



#------------------------------------------------------------------------------#


  def _LinearTerms_pi(self):
    return self._UniaxialAnisotropyTerm() + self._MaxwellTerm() \
      + self._DzyaloshinskiiMoriyaTerm()


#------------------------------------------------------------------------------#


  def _NonLinearTerms_Pi(self):
    return self._SpintronicsTerm()


#------------------------------------------------------------------------------#


  def _UniaxialAnisotropyTerm(self):
    return -self._C_ani * self._UniaxialAnisotropy(self._Mag.m)


#------------------------------------------------------------------------------#


  def _MaxwellTerm(self):
    return self._Maxwell(self._Mag)


#------------------------------------------------------------------------------#


  def _DzyaloshinskiiMoriyaTerm(self):
    return -0.5 * self._C_dm \
      * self._DzyaloshinskiiMoriya(self._Mag.m, self._Mag.Dm)


#------------------------------------------------------------------------------#


  def _SpintronicsTerm(self):
    return self._C_st * self._Spintronics(self._Mag.m, self._Mag.Dm)


#------------------------------------------------------------------------------#

  def _AppliedTerm(self):
    return self._Applied()


#------------------------------------------------------------------------------#


  def _Mass(self):
    from . import _bilinearForms
    return _bilinearForms.Mass(self._psi, self._phi)


#------------------------------------------------------------------------------#


  def _Stiff(self):
    from . import _bilinearForms
    return _bilinearForms.Stiff(self._Dpsi, self._Dphi)


#------------------------------------------------------------------------------#


  def _Skew(self):
    from . import _bilinearForms
    return _bilinearForms.Skew(self._psi, self._phi, self._Mag.m)


#------------------------------------------------------------------------------#


  def _Exchange(self):
    from . import _linearForms
    return _linearForms.Exchange(self._Dphi, self._Mag.Dm)


#------------------------------------------------------------------------------#


  def _UniaxialAnisotropy(self, gf):
    from . import _linearForms
    if(self.parameters.K == 0.0):
      return self._ZeroCF()

    return _linearForms.UniaxialAnisotropy(self._phi, self._easy_cf \
      , gf)


#------------------------------------------------------------------------------#


  def _Maxwell(self, gf_obj):
    import sys
    if(self.parameters.maxwellCoupling == "none"):
      return self._ZeroCF()

    if(self.parameters.maxwellCoupling == "magnetostaticField"):
      return self._Magnetostatic(gf_obj)
    elif(self.parameters.maxwellCoupling == "strayField"):
      return self._Stray(gf_obj)
    elif(self.parameters.maxwellCoupling == "oerstedField"):
      return self._Oersted(gf_obj)
    elif(self.parameters.maxwellCoupling == "eddyCurrent"):
      print("Eddy Currents TOECDO")
      sys.exit()
    else:
      print("This option is not implemented.")
      sys.exit()


#------------------------------------------------------------------------------#


  def _DzyaloshinskiiMoriya(self, gf, Dgf):
    import sys
    if(self.parameters.dmCoupling == "none"):
      return self._ZeroCF()

    if(self.parameters.dmCoupling == "bulk"):
      return self._BulkDMI(gf, Dgf)
    elif(self.parameters.dmCoupling == "interfacial"):
      return self._InterfacialDMI(gf, Dgf)
    else:
      print("This option is not implemented.")
      sys.exit()


#------------------------------------------------------------------------------#


  def _Magnetostatic(self, gf_obj):
    return self._Stray(gf_obj) \
      + self._Oersted(gf_obj)


#------------------------------------------------------------------------------#


  def _Stray(self, gf_obj):
    from . import _linearForms
    return _linearForms.Stray(self._phi, gf_obj.hs.cf)


#------------------------------------------------------------------------------#


  def _Oersted(self, gf_obj):
    from . import _linearForms
    return _linearForms.Oersted(self._phi, gf_obj.hc.cf)


#------------------------------------------------------------------------------#


  def _BulkDMI(self, gf, Dgf):
    from . import _linearForms
    return _linearForms.BulkDMI(self._phi, gf, self._Dphi, Dgf)


#------------------------------------------------------------------------------#


  def _InterfacialDMI(self, gf, Dgf):
    from . import _linearForms
    return _linearForms.InterfacialAxisDMI(self._phi, gf, self._Dphi, Dgf \
      , self.parameters.dmi_axis)


#------------------------------------------------------------------------------#

  
  def _Spintronics(self, gf, Dgf):
    import sys
    if(self.parameters.spintronicsCoupling == "none"):
      return self._ZeroCF()

    if(self.parameters.spintronicsCoupling == "zhangLi"):
      return self._ZhangLi(gf, Dgf)
    elif(self.parameters.spintronicsCoupling == "slonczewski"):
      return self._Slonczewski(gf)
    else: 
      print("This option is not implemented.")
      sys.exit()


#------------------------------------------------------------------------------#


  def _ZhangLi(self, gf, Dgf):
    from . import _linearForms
    return _linearForms.ZhangLiFirst(self._phi, self._j_cf, Dgf, gf) \
      + self.parameters.xi \
      * _linearForms.ZhangLiSecond(self._phi, self._j_cf, Dgf)


#------------------------------------------------------------------------------#


  def _Slonczewski(self, gf):
    from . import _linearForms
    return _linearForms.SlonczewskiFirst(self._phi, self._j_cf \
      , self.parameters.P, self._p_cf, gf, gf)


#------------------------------------------------------------------------------#


  def _H_proj(self, x, y, z):
    x, y, z = self._Q_axisT([x, y, z])

    sigma = 1 if z > 0 else -1
        
    scale = 1.0 + abs(z)

    R0 = 1.0 - x*x / scale
    R12 = - x*y / scale 
    R3 = 1.0 - y*y / scale
    R4 = - x*(z + sigma) / scale
    R5 = - y*(z + sigma) / scale
    
    return self._Q_axisT([R0, R12, R12, R3, R4, R5])
  
  
#------------------------------------------------------------------------------#


  def _R_proj(cls, x, y, z):
    epsilon = 1.0e-12

    x, y, z = self._Q_axisT([x, y, z])
        
    scale = 1.0+z

    R0 = 1.0 
    R12 = 0.0 
    R3 = 1.0 
    R4 = -x
    R5 = -y
    
    if ( abs(scale) > epsilon ):
      R0 -= x*x/scale
      R12 -= x*y/scale
      R3 -= y*y/scale

    return self._Q_axisT([R0, R12, R12, R3, R4, R5])
  
  
#------------------------------------------------------------------------------#


  def _Compute_Q(self):
    import scipy.sparse as sp
    from . import _linearForms
    import numpy as np
    N = self._Mag.m.components[0].space.ndof
        
    m_fv = self._Mag.m.vec.FV().NumPy()
    Q_data = np.empty(6*N, dtype=float)
    D = np.empty(6*N, dtype=float)

    for j in range(N):
      D[6*j:6*(j+1)] \
        = self._Q_proj(m_fv[j], m_fv[j+N], m_fv[j+2*N])
   
    E = D.reshape((6,N), order='F')
    Q_data[:2*N] = E[:2,:].reshape((2*N,), order='F')
    Q_data[2*N:4*N] = E[2:4,:].reshape((2*N,), order='F')
    Q_data[4*N:] = E[4:,:].reshape((2*N,), order='F')

    tmp = np.arange(2*N)
    Q_rows = np.append(tmp, [tmp, tmp])
    Q_cols = np.repeat(np.arange(3*N), 2)
        
    self._Q = sp.csr_matrix( (Q_data, (Q_rows, Q_cols)), shape=(2*N,3*N) )   
  

#------------------------------------------------------------------------------#


################################################################################
