###  TPS2+AB ###
from . import tps2


################################################################################


class TPS2AB(tps2.TPS2):
  # description of TPS2


#------------------------------------------------------------------------------#

  
  def __init__(self, geometry, parameters = None, output = None, numthreads = 1):
    
    tps2.TPS2.__init__(self, geometry, parameters, output, numthreads)

    self._nowImplicit = True


#------------------------------------------------------------------------------#


  def GetReady(self):
    self._tsm_new = 1.0 + self._theta
    self._tsm_old = -self._theta
    
    tps2.TPS2.GetReady(self)
    self._nowImplicit = True


#------------------------------------------------------------------------------#


  def _SetRHS_3D(self):
    if self._nowImplicit:
      tps2.TPS2._SetRHS_3D(self)
    else:
      self._RHS_3D = self._ab_f.vec.FV().NumPy()


#------------------------------------------------------------------------------#


  def _SetUpAuxiliaryFunctions(self):
    tps2.TPS2._SetUpAuxiliaryFunctions(self)

    self._SetUpOldMagnetization()


#------------------------------------------------------------------------------#

  
  def _SetUpOldMagnetization(self):
    from commix._gridFunctions import Magnetization
#    if self.parameters.spintronicsCoupling == "slonczewski":
      # for _Mag_old, the OerstedField is never computed.
      # of j_prime = None, it's also not needed
      # TODO: OERSTED take care of this in the future, when Oersted is supported
      # jprime_cf = self._jprime_cf*self._p_cf if self._jprime_cf != None else \
      # self._jprime_cf
#    else:
#      self._Mag_old = Magnetization(self._geometry, self._V, self._X \
#        , self._Mag \
#        , self.parameters.maxwellCoupling, self._solvetol, self._j_cf )
    jprime_cf = None
    self._Mag_old = Magnetization(self._geometry, self._V, self._X \
      , self._Mag \
      , self.parameters.maxwellCoupling, self._solvetol, jprime_cf)


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):

    tps2.TPS2._UpdateMaxwellCoupling(self)

    if self.parameters.spintronicsCoupling == "slonczewski":
      self._Mag_old.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
        , self._solvetol, self._j_cf * self._p_cf)
    else:
      self._Mag_old.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
        , self._solvetol, self._j_cf)


#------------------------------------------------------------------------------#


  def _StepFollowUp(self):
    
    tps2.TPS2._StepFollowUp(self)

    self._nowImplicit = False


#------------------------------------------------------------------------------#


  def _StepUpdate(self):
    self._Mag_old.SetFromOther(self._Mag)

    tps2.TPS2._StepUpdate(self)


#------------------------------------------------------------------------------#


  def _SetUpSystems(self):
    from ngsolve import SymbolicLFI, LinearForm

    tps2.TPS2._SetUpSystems(self)

    #lf
    rhs_ab_cf = self._ExchangeTerm()
    rhs_ab_cf += self._ab_LinearTerms_pi()
    rhs_ab_cf += self._ab_NonLinearTerms_Pi()
    rhs_ab_cf += self._AppliedTerm()

    self._ab_f = LinearForm(self._X)
    self._ab_f += SymbolicLFI(rhs_ab_cf.Compile(realcompile=True, wait=True))



#------------------------------------------------------------------------------#


  def _Compute_v(self):
    import scipy.sparse.linalg
    if(self._nowImplicit):
      tps2.TPS2._Compute_v(self)
      return

    self._solution, succ = scipy.sparse.linalg.gmres(self._LHS, self._RHS \
      , x0=self._guess, tol=self._solvetol, maxiter=4000 \
      , M=self._preconditioner, callback=self._callback)

    self._Mag.v.gf.vec.FV().NumPy()[:] = self._Q.transpose().dot(self._solution)

    self._UpdateGuess()
    

#------------------------------------------------------------------------------#


  def _AssembleRHS(self):
    self._Assemble_f_ab()


#------------------------------------------------------------------------------#


  def _Assemble_f_ab(self):
    self._ab_f.Assemble()


#------------------------------------------------------------------------------#


  def _ab_LinearTerms_pi(self):
    return self._tsm_new * self._LinearTerms_pi() \
      + self._tsm_old * self._OldLinearTerms_pi()


#------------------------------------------------------------------------------#


  def _ab_NonLinearTerms_Pi(self):
    return self._NonLinearTerms_Pi() \
      + (self._tsm_new - 1.0) * self._NowDerivNonLinearTerms() \
      + self._tsm_old * self._OldDerivNonLinearTerms()


#------------------------------------------------------------------------------#


  def _OldLinearTerms_pi(self):
    return self._OldUniaxialAnisotropyTerm() \
      + self._OldMaxwellTerm() \
      + self._OldDzyaloshinskiiMoriyaTerm()


#------------------------------------------------------------------------------#


  def _NowDerivNonLinearTerms(self):
    return self._NowDerivSpintronicsTerm()


#------------------------------------------------------------------------------#


  def _OldDerivNonLinearTerms(self):
    return self._OldDerivSpintronicsTerm()


#------------------------------------------------------------------------------#


  def _OldUniaxialAnisotropyTerm(self):
    return -self._C_ani * self._OldUniaxialAnisotropy()


#------------------------------------------------------------------------------#


  def _OldMaxwellTerm(self):
    return self._OldMaxwell()


#------------------------------------------------------------------------------#


  def _OldDzyaloshinskiiMoriyaTerm(self):
    return -0.5 * self._C_dm * self._OldDzyaloshinskiiMoriya()


#------------------------------------------------------------------------------#


  def _NowDerivSpintronicsTerm(self):
    return self._C_st * self._NowDerivSpintronics()


#------------------------------------------------------------------------------#


  def _OldDerivSpintronicsTerm(self):
    return self._C_st * self._OldDerivSpintronics()


#------------------------------------------------------------------------------#


  def _OldUniaxialAnisotropy(self):
    return self._UniaxialAnisotropy(self._Mag_old.m)


#------------------------------------------------------------------------------#


  def _OldMaxwell(self):
    # TOECDO Achtung funktioniert so nur fuer Streufeld
    return self._Maxwell(self._Mag_old)


#------------------------------------------------------------------------------#


  def _OldDzyaloshinskiiMoriya(self):
    return self._DzyaloshinskiiMoriya(self._Mag_old.m, self._Mag_old.Dm)


#------------------------------------------------------------------------------#


  def _NowDerivSpintronics(self):
    return self._DerivSpintronics(self._Mag.m, self._Mag.m \
      , self._Mag.Dm, self._Mag.Dm)


#------------------------------------------------------------------------------#


  def _OldDerivSpintronics(self):
    return self._DerivSpintronics(self._Mag.m, self._Mag_old.m \
      , self._Mag.Dm, self._Mag_old.Dm)


#------------------------------------------------------------------------------#


################################################################################
